"""
Normalizer

....................................................................................................
MIT License

Copyright (c) 2021-2023 AUT Iran, Mohammad H Forouhesh
Copyright (c) 2021-2022 MetoData.ai, Mohammad H Forouhesh
....................................................................................................
This Module contains the implementation and encapsulation for Text Normalizer, this functionality
helps with detecting
half-spaces.
"""
import itertools
from re import sub
import os
from typing import Dict, List, Generator

from crf_pos.api import downloader
from crf_pos.normalization.tokenizer import clean_text


class Normalizer:
    """
    A native persian text normalizer to help detecting half-spaces.
    """
    def __init__(self, downloading: bool = False) -> None:
        self.dir_path = os.path.dirname(
            os.path.dirname(
                os.path.dirname(
                    os.path.realpath(__file__)))) + "/"
        if downloading: self.get_resources()
        self.corrections = self.load_dictionary(self.dir_path + 'model/normalizer/corrections.txt')

    def get_resources(self) -> None:
        """
        A tool to download required resources over internet.
        :return:    None.
        """
        load_dir = 'https://github.com/MohammadForouhesh/crf-pos-persian/releases/download/v2.0.0.alpha/corrections.txt'
        save_dir = self.dir_path + '/model/normalizer/'
        os.makedirs(save_dir, exist_ok=True)
        downloader(path=load_dir, save_path=save_dir + 'corrections.txt', mode='wb')

    @staticmethod
    def load_dictionary(file_path: str) -> Dict[str, str]:
        """
        A static method that read a file and return a keyed dictionary of the contents of the file.
        :param file_path:   The path (str) to the resource file.
        :return:            A dictionary of the contents of the file.
        """
        dictionary = {}
        with open(file_path, 'r', encoding='utf-8') as file:
            lines = file.readlines()
            for words in lines:
                word = words.replace('\ufeff', '').split(' ')
                dictionary[word[0].strip()] = sub('\n', '', word[1].strip())
        return dictionary

    @staticmethod
    def space_correction(text: str) -> str:
        """
        A tool to help with rule-based half space correction
        :param text:        The input text (str).
        :return:            The half-spaced corrected text (str).
        """
        text = sub(r'^(بی|می|نمی)( )', r'\1‌', text)
        text = sub(r'( )(می|نمی|بی)( )', r'\1\2‌', text)
        pattern = r'( )(هایی|ها|های|ایی|هایم|هایت|هایش|هایمان|هایتان|هایشان|ات|ان|ین' \
                  r'|انی|بان|ام|ای|یم|ید|اید|اند|بودم|بودی|بود|بودیم|بودید|بودند|ست)( )' \
                  r'( )(طلبان|طلب|گرایی|گرایان|شناس|شناسی|گذاری|گذار|گذاران|شناسان|گیری|پذیری|بندی|آوری|سازی|' \
                  r'بندی|کننده|کنندگان|گیری|پرداز|پردازی|پردازان|آمیز|سنجی|ریزی|داری|دهنده|آمیز|پذیری' \
                  r'|پذیر|پذیران|گر|ریز|ریزی|رسانی|یاب|یابی|گانه|گانه‌ای|انگاری|گا|بند|رسانی|دهندگان|دار)( )'
        text = sub(pattern, r'‌\2\3', text)
        return sub(r'( )(شده|نشده)( )', r'‌\2‌', text)

    @staticmethod
    def window_sampling(tokens: List[str], window_length: int) -> Generator[str, None, None]:
        if len(tokens) < window_length: yield ' '.join(tokens)
        while True:
            try:                yield ' '.join([tokens.pop(0)] + [tokens[_] for _ in range(window_length - 1)])
            except IndexError:  break

    def vector_mavericks(self, text: str, window_length: int) -> Generator[str, None, None]:
        iter_sample = iter(self.window_sampling(text.replace('\u200c', ' ').split(), window_length))
        for word in iter_sample:
            try:
                yield self.corrections[word.replace(' ', '')]
                for ind in range(0, window_length - 1):   next(iter_sample, None)
            except:
                try:    yield self.corrections[word.split()[0]]
                except: yield word.split()[0]

    def moving_mavericks(self, text: str, scope: int = 4) -> Generator[str, None, None]:
        yield self.vector_mavericks(text, scope)
        if scope > 1: yield from self.moving_mavericks(text, scope - 1)

    def collapse_mavericks(self, text: str) -> str:
        mavericks_cascades = list(map(lambda item: ' '.join(item), self.moving_mavericks(text)))
        return sorted(mavericks_cascades, key=lambda item: item.count('\u200c'))[-1]

    def normalize(self, text: str, new_line_elimination: bool = False) -> str:
        """
        Normalizer function, it is the main function in this class.
        :param text:        The input text.
        :param new_line_elimination: A boolean, controls whether to use another character for
                                     half-space.
        :return:            A normalized text.
        """
        cleansed_string = clean_text(text, new_line_elimination).strip()
        return self.space_correction(self.collapse_mavericks(cleansed_string)).strip()



"""
API

....................................................................................................
MIT License

Copyright (c) 2021-2023 AUT Iran, Mohammad H Forouhesh
Copyright (c) 2021-2022 MetoData.ai, Mohammad H Forouhesh
....................................................................................................
This module contains tools to download resources over http connections.
supported http links are:
    - https://github.com/MohammadForouhesh/crf-pos-persian/releases/download/v2.0.0.alpha/UPC_full_model_wapiti
    - https://github.com/MohammadForouhesh/crf-pos-persian/releases/download/v2.0.0.alpha/perpos.model
    - https://raw.githubusercontent.com/MohammadForouhesh/Parsivar/master/parsivar/resource/normalizer/model/normalizer/Dic1_new.txt
    - https://raw.githubusercontent.com/MohammadForouhesh/Parsivar/master/parsivar/resource/normalizer/model/normalizer/Dic2_new.txt
    - https://raw.githubusercontent.com/MohammadForouhesh/Parsivar/master/parsivar/resource/normalizer/model/normalizer/Dic3_new.txt
"""

import os
from typing import Union

import requests


def downloader(path: str, save_path: str, mode: str) -> Union[int, None]:
    """
    A tool to download and save files over https.
    supported http links are:
    - https://github.com/MohammadForouhesh/crf-pos-persian/releases/download/v2.0.0.alpha/UPC_full_model_wapiti
    - https://github.com/MohammadForouhesh/crf-pos-persian/releases/download/v2.0.0.alpha/perpos.model
    - https://raw.githubusercontent.com/MohammadForouhesh/Parsivar/master/parsivar/resource/normalizer/model/normalizer/Dic1_new.txt
    - https://raw.githubusercontent.com/MohammadForouhesh/Parsivar/master/parsivar/resource/normalizer/model/normalizer/Dic2_new.txt
    - https://raw.githubusercontent.com/MohammadForouhesh/Parsivar/master/parsivar/resource/normalizer/model/normalizer/Dic3_new.txt

    :param path:        The path to the desired file.
    :param save_path:   The intended storage path.
    :param mode:        The mode that it should be stored.
    :return:            If the file exists, it returns 0 (int), otherwise nothing would be returned.
    """
    if os.path.isfile(save_path): return 0
    try:
        model_bin = requests.get(path, allow_redirects=True)
        with open(save_path, mode) as resource:
            resource.write(model_bin.content)
    except Exception:
        raise Exception('not a proper webpage')



"""
Wapiti Classifier

....................................................................................................
MIT License

Copyright (c) 2021-2023 AUT Iran, Mohammad H Forouhesh
Copyright (c) 2021-2022 MetoData.ai, Mohammad H Forouhesh
....................................................................................................
This module sole purpose is abstraction, it contains a meta class for Wapiti and CRF classifier.
"""

from typing import Union, List, Any, Tuple, Generator
from crf_pos.normalization.normalizer import Normalizer


class MetaTagger:
    """
    Part-of-Speech taggers meta class abstraction.
    """
    def __init__(self) -> None:
        self.tagger = None
        self.norm = Normalizer(downloading=True)

    def __getitem__(self, item: Union[list, str]) -> List[Tuple[str, str]]:
        """
        Generic getitem, used across the children objects.
        :param item:    An input text/list, the preferred input type is string.
        :return:        A List of tuples including a word and its part of speech.
        """
        if isinstance(item, str):   item = self.norm.normalize(item).split()
        return self.parse([item])[0]

    ## ToDo get_resources.

    def parse(self, token_list: List[str]) -> List[List[Tuple[Any, Any]]]:
        """
        An abstract method, to be overwritten by its descendants.
        :param token_list:  A list of tokens (strings)
        :return:            A list of list of extracted part of speeches and their related tokens.
        """
        pass

    @staticmethod
    def zip_vector(iterable: zip) -> Generator[List[zip], None, None]:
        """
        A function that combine to iterables elements together.
        :param iterable:    A zip that contains the zipped format of two iterables.
        :return:            A Generator containing a list of tuples that consist of elements of
                            each iterable.
        """
        for key, item in iterable:
            yield list(zip(key, item))



"""
Utils

....................................................................................................
MIT License

Copyright (c) 2021-2023 AUT Iran, Mohammad H Forouhesh
Copyright (c) 2021-2022 MetoData.ai, Mohammad H Forouhesh
....................................................................................................
This module contains various tools and one-line functions for Part-of-Speech tagging.
"""
from typing import Generator, Dict, Any

import string
import re

token2features = lambda item: [word2features(item, ind) for ind in range(len(item))]
sent2labels = lambda item: [postag for token, postag in item]
sent2tokens = lambda item: [token for token, postag in item]
template = lambda word: "".join([(lambda item: "x" if not item in "آایو" else "a")(char)
                                 for char in word])
isdigit = lambda word: all(map(lambda char: char in "۱۲۳۴۵۶۷۸۹۰1234567890.", word))
is_all_latin = lambda item: bool(len(re.sub('[a-zA-Z]*', '', item)) == 0)
remove_after_underline = lambda item: item[:item.find('_')] if '_' in item else item


def ngram(text: str, length: int = 2) -> Generator[str, None, None]:
    """
    function for detecting n-grams.
    :param text:    Input text, it is in the form of a sentence and it is a string.
    :param length:  The length at which we want to extract n-grams.
    :return:        A Generator, it yields words pattern.
    """
    for i in range(len(text) - 1):
        yield 'word[' + str(i) + ":" + str(i + length) + "]", text[i:i + length]


def word2features(text: str, index: int) -> Dict[str, Any]:
    """
    A feature extraction tool that helps with extraction of different useful information
    within the given text.
    :param text:    The input text. A string.
    :param index:   The place of the word at which we want to move the window back and forth.
    :return:        A dictionary containing extracted features.
    """
    word = text[index]
    features = {
        'B': 1.0,
        'W': word,
        'P': word in string.punctuation,
        'T': template(word),
        'D(W)': isdigit(word),
    }
    for length in range(max(4 + 1, len(word)) + 1):
        for k, v in ngram(word, length=length):
            features[k] = v
    if index > 0:
        word = text[index - 1][0]
        features.update({
            '-1W[-3': word[-3:],
            '-1W[-2': word[-2:],
            '-1W[-1': word[-1:],
            '-1W': word,
            '-1W0W': word + text[index],
            '-1P': word in string.punctuation,
            '-1T': template(word)
        })
    else:
        features['BOS'] = True
    if index > 1:
        word = text[index - 2][0]
        features.update({
            '-2W[-3': word[-3:],
            '-2W[-2': word[-2:],
            '-2W[-1': word[-1:],
            '-2P': word in string.punctuation,
            '-2T': template(word)
        })

    if index < len(text) - 2:
        word = text[index + 2][0]
        features.update({
            '+2W[-1': word[-1:],
            '+2W[-2': word[-2:],
            '+2W': word,
            '+2P': word in string.punctuation,
            '+2T': template(word)
        })
    if index < len(text) - 1:
        word = text[index + 1][0]
        features.update({
            '+1W[-1': word[-1:],
            '+1W': word,
            '+1W0W': word + text[index],
            '+1W[-2': word[-2:],
            '+1:P': word in string.punctuation,
            '+1:T': template(word)
        })
    else:
        features['EOS'] = True
    if 0 < index < len(text) - 1:   features['-1W/+1W'] = text[index + 1][0] + "/" + text[index - 1][0]
    return features



"""
CRF Classifier

....................................................................................................
MIT License

Copyright (c) 2021-2023 AUT Iran, Mohammad H Forouhesh
Copyright (c) 2021-2022 MetoData.ai, Mohammad H Forouhesh
....................................................................................................
This Module contains the implementation and encapsulation for Conditional Random Field classifier.
"""

from typing import List, Tuple, Any
from crf_pos.pos_tagger.meta_tagger import MetaTagger
from crf_pos.pos_tagger.utils import token2features
import pickle


class CrfPosTagger(MetaTagger):
    """
    Wapiti Part-of-Speech tagger encapsulation.
    """
    def __init__(self, model_path) -> None:
        super().__init__()
        with open(model_path, 'rb') as resource:
            self.tagger = pickle.load(resource)

    def parse(self, token_list: List[str]) -> List[List[Tuple[Any, Any]]]:
        """
        Primary function that overwrite the same method from the super class. This function is
        responsible for the logic behind the model.
        :param token_list:  A list of tokens (strings)
        :return:            A list of list of extracted part of speeches and their related tokens.
        """
        y_pred = self.tagger.predict([token2features(token) for token in token_list])
        return list(super().zip_vector(zip(token_list, y_pred)))



"""
Wapiti Classifier

....................................................................................................
MIT License

Copyright (c) 2021-2023 AUT Iran, Mohammad H Forouhesh
Copyright (c) 2021-2022 MetoData.ai, Mohammad H Forouhesh
....................................................................................................
This module contains the implementation and encapsulation for Wapiti classifier.
"""

from typing import List, Any, Tuple
from wapiti import Model
from crf_pos.pos_tagger.meta_tagger import MetaTagger
from crf_pos.pos_tagger.utils import remove_after_underline


class WapitiPosTagger(MetaTagger):
    """
    Wapiti Part-of-Speech tagger encapsulation.
    """
    def __init__(self, model_path: str = 'model/UPC_full_model_wapiti') -> None:
        super().__init__()
        self.tagger = Model(model=model_path)

    def parse(self, token_list: List[str]) -> List[List[Tuple[Any, Any]]]:
        """
        Primary function that overwrite the same method from the super class. This function is
        responsible for the logic behind the model.
        :param token_list:  A list of tokens (strings)
        :return:            A list of list of extracted part of speeches and their related tokens.
        """
        sent_line = "\n".join(token_list[0])
        postags = self.tagger.label_sequence(sent_line).decode('utf-8').strip().split('\n')
        return list(super().zip_vector(zip(token_list, [list(map(remove_after_underline, postags))])))



import sys
import distutils.util
from setuptools import setup, find_packages
from pathlib import Path

PACKAGES = find_packages()
COMPILE_OPTIONS = {
    "msvc": ["/Ox", "/EHsc"],
    "other": ["-O3", "-Wno-strict-prototypes", "-Wno-unused-function"],
}
COMPILER_DIRECTIVES = {
    "language_level": -3,
    "embedsignature": True,
    "annotation_typing": False,
}
LINK_OPTIONS = {"msvc": [], "other": []}


def is_new_osx():
    """Check whether we're on OSX >= 10.10"""
    name = distutils.util.get_platform()
    if sys.platform != "darwin":
        return False
    elif name.startswith("macosx-10"):
        minor_version = int(name.split("-")[1].split(".")[1])
        if minor_version >= 7:
            return True
        else:
            return False
    else:
        return False


if is_new_osx():
    COMPILE_OPTIONS["other"].append("-stdlib=libc++")
    LINK_OPTIONS["other"].append("-lc++")
    LINK_OPTIONS["other"].append("-nodefaultlibs")


def clean(path):
    for path in path.glob("**/*"):
        if path.is_file() and path.suffix in (".so", ".cpp"):
            print(f"Deleting {path.name}")
            path.unlink()


def setup_package():
    root = Path(__file__).parent

    setup(
        name='crf_pos',
        packages=PACKAGES,
        version='2.0.0',
        url='https://github.com/MohammadForouhesh/crf-pos-persian',
        license='MIT',
        author='MohammadForouhesh',
        author_email='Mohammadh.Forouhesh@gmail.com',
        description='Persian Part-of-Speech tagger framework',
        package_data={"": ["*.pyx", "*.pxd", "*.pxi", "*.cu"]},
    )


if __name__ == '__main__':
    setup_package()



"""
Tokenizer

....................................................................................................
MIT License

Copyright (c) 2021-2023 AUT Iran, Mohammad H Forouhesh
Copyright (c) 2021-2022 MetoData.ai, Mohammad H Forouhesh
....................................................................................................
This Module contains the tools and one-line functions used in tokenization process.
"""

import re
import string
from typing import List


tokenize_words = lambda text: [item.strip("\u200c") for item in text.strip().split()
                               if len(item.strip("\u200c")) != 0]
add_tab = lambda text: str(" " + text.group().strip(' ').strip('\n') + "\t\t")
triplets2tagged_pairs = lambda iob_sent: [((word, pos), chunk) for word, pos, chunk in iob_sent]
add_space = lambda text: str(" " + text.group().strip(' ') + " ")


def tokenize_sentences(text: str) -> List[str]:
    """
    A sentencizer function, split text into sentence.
    :param text:    A text.
    :return:        Splatted sentences in the format of list of strings.
    """
    text = pattern_matching(text)
    nums_list = re.findall(r"[-+]?\d*\.\d+|\d+", text)
    for number in nums_list:
        pattern = 'floatingpointnumber'
        text = re.sub(pattern, number, text, 1)

    return [item for item in text.split('\t\t') if len(item) != 0]


def pattern_matching(text: str) -> str:
    """
    A regex based pattern matching function to remove the unwanted characters in the text.
    :param text:    The input text (str).
    :return:        A text without the patterns.
    """
    patterns = re.compile(pattern="["
                                  r"[-+]?\d*\.\d+|\d+"
                                  r"([!\.\?؟]+)[\n]*"
                                  r":\n"
                                  r";\n"
                                  r'؛\n'
                                  r'[\n]+'
                                  "]+", flags=re.UNICODE)
    return str(patterns.sub(r'', text))


def clean_text(text_doc: str, new_line_elimination: bool) -> str:
    """
    A preprocessing function to delete the punctuations.
    :param text_doc:    The input text.
    :param new_line_elimination: A boolean that controls the usage of half-space
    :return:            Removed punctuation text.
    """
    punctuations = r')(}{:؟!،؛»«.' + r"/<>?.,:;"
    punctuations = '[' + punctuations + string.punctuation + ']'
    punctuations = punctuations.replace("@", "")

    text_doc.strip()

    pattern = r"[-+]?\d*\.\d+|\d+"
    nums_list = re.findall(pattern, text_doc)
    new_string = re.sub(pattern, 'floatingpointnumber', text_doc)

    pattern = r'\s*' + punctuations + '+' + r'\s*'
    new_string = re.sub(pattern, add_space, new_string)

    pattern = r'[\n]+'
    if new_line_elimination:  new_string = re.sub(pattern, " ", new_string)

    pattern = r'[^' + r")(}{:؟!-،؛»«.@$&%" + r"/<>?.,:;" + r"a-zA-Z0-9" + 'آ-ی' + '‌' + '\d\s:]'
    new_string = re.sub(pattern, '', new_string)

    pattern = r'[ ]+'
    new_string = re.sub(pattern, ' ', new_string)

    for number in nums_list:
        pattern = 'floatingpointnumber'
        new_string = re.sub(pattern, number, new_string, 1)
    return new_string
